/**
    Copyright (C) 2001, 2002  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <semaphore.h>
#include <tcstring.h>
#include <pthread.h>
#include <tclist.h>
#include <tchash.h>
#include "tc2.h"
#include "eventq.h"

static tclist_t *queues;
static tchash_table_t *names;
static pthread_mutex_t dlock;
static unsigned long last_event = 0;

struct eventq {
    unsigned long last;
    tclist_t *q;
    sem_t new_event;
    tclist_t *send, *recv;
    void *(*sndfun)(void *);
};

struct namedq {
    char *name;
    tclist_t *send, *recv;
};

static struct namedq *findn(char *name, int create);
static int sendq(struct eventq *q, void *data);
static int sendn(struct namedq *nq, void *data);
static int e_cmp(const void *, const void *);
static int n_cmp(const void *a, const void *b);
int eq_detach(eventq_t type, char *name, int mode);
static void free_q(struct eventq *);

static void
free_nq(void *p)
{
    struct namedq *nq = p;
    free(nq->name);
    if(nq->send)
	tclist_destroy(nq->send, NULL);
    if(nq->recv)
	tclist_destroy(nq->recv, NULL);
}

/* Initialization function */
extern int
event_init(char *opt)
{
#ifdef EQ_DEBUG
    fprintf(stderr, "Event queue initializing...\n");
#endif
    names = tchash_new(128, TC_LOCK_SLOPPY, 0);
    queues = tclist_new(TC_LOCK_SLOPPY);
    pthread_mutex_init(&dlock, NULL);
    return 0;
}

/* Shutdown function */
extern int
event_shdn(void)
{
    int qs = 0;
    void qfree(void *q){
	qs++;
	free_q(q);
    }
#ifdef EQ_DEBUG
    fprintf(stderr, "Event queue shutting down...\n");
#endif

    tclist_destroy(queues, qfree);
    tchash_destroy(names, NULL);
    pthread_mutex_destroy(&dlock);
    return 0;
}

extern eventq_t
register_event_type(void *(*sndfun)(void *))
{
    struct eventq *q;

    q = calloc(1, sizeof(struct eventq));
    q->last = 0;
    q->q = tclist_new(TC_LOCK_SLOPPY);
    sem_init(&q->new_event, 0, 0);
    q->sndfun = sndfun;
    tclist_push(queues, q);
    return q;
}

extern int
unregister_event_type(eventq_t q)
{
    tclist_delete(queues, q, NULL, (tcfree_fn) free_q);
    return 0;
}

static void
free_q(struct eventq *q)
{
    struct namedq *nq;
    /* Detach from any attached names. */
    if(q->send != NULL){
	while(q->send != NULL && (nq = tclist_shift(q->send)) != NULL){
	    eq_detach(q, nq->name, EVENTQ_SEND);
	}
/* 	tclist_free(q->send); */
    }
    if(q->recv != NULL){
	while(q->recv != NULL && (nq = tclist_shift(q->recv)) != NULL){
	    eq_detach(q, nq->name, EVENTQ_RECV);
	}
/* 	tclist_free(q->recv); */
    }

    sem_destroy(&q->new_event);
    tclist_destroy(q->q, NULL);
    free(q);
}


extern int
eq_attach(eventq_t q, char *name, int mode)
{
    struct namedq *nq;

    nq = findn(name, 1);

    if(mode == EVENTQ_SEND){
	if(nq->send == NULL)
	    nq->send = tclist_new(TC_LOCK_SLOPPY);
	tclist_search(nq->send, q, NULL, e_cmp);
	if(q->send == NULL)
	    q->send = tclist_new(TC_LOCK_SLOPPY);
	tclist_search(q->send, nq, NULL, n_cmp);
    } else if(mode == EVENTQ_RECV){
	if(nq->recv == NULL)
	    nq->recv = tclist_new(TC_LOCK_SLOPPY);
	tclist_search(nq->recv, q, NULL, e_cmp);
	if(q->recv == NULL)
	    q->recv = tclist_new(TC_LOCK_SLOPPY);
	tclist_search(q->recv, nq, NULL, n_cmp);
    }
    return 0;
}	


extern int
eq_detach(eventq_t q, char *name, int mode)
{
    struct namedq *nq;

    if((nq = findn(name, 0)) == NULL){
	return -1;
    }

    if(mode == EVENTQ_SEND){
	if(nq->send != NULL){
	    tclist_delete(nq->send, q, NULL, NULL);
	    if(tclist_items(nq->send) == 0){
		tclist_free(nq->send);
		nq->send = NULL;
	    }
	}
	if(q->send != NULL){
	    tclist_delete(q->send, nq, n_cmp, NULL);
	    if(tclist_items(q->send) == 0){
		tclist_free(q->send);
		q->send = NULL;
	    }
	}
    } else if(mode == EVENTQ_RECV){
	if(nq->recv != NULL){
	    tclist_delete(nq->recv, q, NULL, NULL);
	    if(tclist_items(nq->recv) == 0){
		tclist_free(nq->recv);
		nq->recv = NULL;
	    }
	}
	if(q->recv != NULL){
	    tclist_delete(q->recv, nq, n_cmp, NULL);
	    if(tclist_items(q->recv) == 0){
		tclist_free(q->recv);
		q->recv = NULL;
	    }
	}
    }

    if(nq->send == NULL && nq->recv == NULL){
	tchash_delete(names, name, -1, NULL);
	free(nq->name);
	free(nq);
    }

    return 0;
}


static struct namedq *
findn(char *name, int create)
{
    struct namedq *nq = NULL;

    if(tchash_find(names, name, -1, &nq) == 1 && create){
	nq = calloc(1, sizeof(struct namedq));
	nq->name = strdup(name);
	tchash_replace(names, name, -1, nq, NULL);
    }

    return nq;
}


static int
sendq(struct eventq *q, void *data)
{
    int n = 0;
    if(q->send == NULL){
	tclist_push(q->q, q->sndfun? q->sndfun(data): data);
	sem_post(&q->new_event);
	n++;
    } else {
	tclist_item_t *li = NULL;
	struct namedq *nq;
	q->last = last_event;
	while((nq = tclist_next(q->send, &li)) != NULL){
	    n += sendn(nq, data);
	}
    }
    return n;
}


static int
sendn(struct namedq *nq, void *data)
{
    int n = 0;

    if(nq->recv != NULL){
	tclist_item_t *li = NULL;
	struct eventq *q;
	while((q = tclist_next(nq->recv, &li)) != NULL){
	    if(q->last < last_event){
		q->last = last_event;
		n += sendq(q, data);
	    }
	}
    }

    return n;
}


extern int
send_event(eventq_t q, void *data)
{
    int n;
    if(q->send != NULL){
	pthread_mutex_lock(&dlock);
	last_event++;
    }
    n = sendq(q, data);
    if(q->send != NULL){
	pthread_mutex_unlock(&dlock);
    }

    return n;
}

extern void *
next_event(eventq_t q)
{
    sem_wait(&q->new_event);
    return tclist_shift(q->q);
}

static int
e_cmp(const void *a, const void *b)
{
    return a - b;
}

static int
n_cmp(const void *a, const void *b)
{
    const struct namedq *e1 = a, *e2 = b;
    return strcmp(e1->name, e2->name);
}
