/**
    Copyright (C) 2001, 2002  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <tcstring.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <tctypes.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <pthread.h>
#include <alloca.h>
#include <tchash.h>
#include <tcalloc.h>
#include <curl/curl.h>
#include <http_tc2.h>

typedef struct _http_private {
    size_t position; /* Bytes read from server */
    CURL *curl;
    char *cachename;
    int cachefd;
    pthread_cond_t rc;
    pthread_mutex_t mtx;
    pthread_t read_th;
} http_private_t;


static tchash_table_t *file_cache;

static void
http_free(void *p)
{
    url_http_t *uh = p;
    http_private_t *hp = uh->private;
    pthread_join(hp->read_th, NULL);
    if(uh->status_str)
	free(uh->status_str);
    if(uh->content_type)
	free(uh->content_type);
    unlink(hp->cachename);
    free(hp->cachename);
    pthread_cond_destroy(&hp->rc);
    pthread_mutex_destroy(&hp->mtx);
    free(hp);
    free(uh);
}


extern int
http_init(char *opt)
{
    file_cache = tchash_new(128, TC_LOCK_SLOPPY, 0);

    if(url_http_conf_cache_dir){
	struct stat st;
	if(stat(url_http_conf_cache_dir, &st) < 0 ||
	   !S_ISDIR(st.st_mode) ||
	   access(url_http_conf_cache_dir, W_OK)){
	    url_http_conf_cache_dir = NULL;
	}
    }

    curl_global_init(CURL_GLOBAL_ALL);

    return 0;
}

extern int
http_stop(void)
{
    curl_global_cleanup();
    tchash_destroy(file_cache, http_free);
    return 0;
}


static size_t
http_hdr(void *buf, size_t size, size_t n, void *s)
{
    url_http_t *uh = s;
    http_private_t *hp = uh->private;
    size_t len = size * n;
    char *hdr = buf;

    hdr[len - 2] = 0;

    pthread_mutex_lock(&hp->mtx);

    if(!strncmp(hdr, "HTTP", 4)){
	char *se;
	uh->status = strtol(hdr + 9, &se, 10);
	uh->status_str = strdup(se);
	hp->position += len;
    } else if(!strncmp(hdr, "Content-Length", 14)){
	uh->size = strtol(hdr + 16, NULL, 0);
    } else if(!strncmp(hdr, "Content-Type", 12)){
	uh->content_type = strdup(hdr + 14);
    }

    pthread_cond_broadcast(&hp->rc);
    pthread_mutex_unlock(&hp->mtx);

    return len;
}

static size_t
http_recv(void *buf, size_t size, size_t n, void *s)
{
    url_http_t *uh = s;
    http_private_t *hp = uh->private;
    size_t len = size * n;

    while(len > 0){
	int l = write(hp->cachefd, buf, len);
	if(l < 0)
	    return -1;
	pthread_mutex_lock(&hp->mtx);
	hp->position += l;
	len -= l;
	pthread_cond_broadcast(&hp->rc);
	pthread_mutex_unlock(&hp->mtx);
    }

    return size * n;
}

static void *
http_fetch(void *p)
{
    url_http_t *uh = p;
    http_private_t *hp = uh->private;

    curl_easy_perform(hp->curl);
    curl_easy_cleanup(hp->curl);

    pthread_mutex_lock(&hp->mtx);
    close(hp->cachefd);
    hp->cachefd = -1;
    pthread_cond_broadcast(&hp->rc);
    pthread_mutex_unlock(&hp->mtx);

    return NULL;
}

static inline void
http_wait(http_private_t *hp, long pos)
{
    pthread_mutex_lock(&hp->mtx);
    while(pos > hp->position && hp->cachefd >= 0){
	pthread_cond_wait(&hp->rc, &hp->mtx);
    }
    pthread_mutex_unlock(&hp->mtx);
}


extern int
http_read(void *data, size_t size, size_t count, url_t *uf)
{
    url_http_t *uh = uf->private;
    size_t pos, bs;
    pos = ftell(uh->cache);
    bs = size * count;
    http_wait(uh->private, pos + bs);
    return fread(data, size, count, uh->cache);
}

extern uint64_t
http_tell(url_t *us)
{
    url_http_t *uh = us->private;

    return ftell(uh->cache);
}

extern int
http_seek(url_t *us, int64_t offset, int how)
{
    url_http_t *uh = us->private;
    http_private_t *hp = uh->private;

    switch(how){
    case SEEK_SET:
	http_wait(hp, offset);
	break;
    case SEEK_CUR:
	http_wait(hp, ftell(uh->cache) + offset);
	break;
    case SEEK_END:
	if(uh->size < 0){
	    pthread_mutex_lock(&hp->mtx);
	    while(hp->cachefd >= 0)
		pthread_cond_wait(&hp->rc, &hp->mtx);
	    pthread_mutex_unlock(&hp->mtx);
	} else {
	    http_wait(hp, uh->size + offset);
	    return fseek(uh->cache, uh->size + offset, SEEK_SET);
	}
	break;
    }

    return fseek(uh->cache, offset, how);
}

extern int
http_close(url_t *uf)
{
    url_http_t *uh = uf->private;
    fclose(uh->cache);
    free(uh);
    tcfree(uf);
    return 0;
}

static int
http_status(url_http_t *u)
{
    int s = u->status;

    switch(s){
    case 0:
    case 100:
    case 101:
    case 300:
    case 301:
    case 302:
    case 303:
    case 305:
    case 307:
	return 0;

    case 200:
    case 203:
    case 206:
	return 1;

    case 201:
    case 202:
    case 304:
    case 306:
    case 400:
    case 411:
	fprintf(stderr, "HTTP: Unexpected status %i %s\n", s, u->status_str);
	return -1;

    case 204:
    case 205:
    case 401:
    case 402:
    case 403:
    case 404:
    case 405:
    case 406:
    case 407:
    case 408:
    case 409:
    case 410:
    case 412:
    case 413:
    case 414:
    case 415:
    case 416:
    case 417:
    case 500:
    case 501:
    case 502:
    case 503:
    case 504:
    case 505:
	fprintf(stderr, "HTTP: %i %s\n", s, u->status_str);
	return -1;
    }

    fprintf(stderr, "HTTP: Unknown status %i %s\n", s, u->status_str);
    return -1;
}

extern url_t *
http_open(char *url, char *mode)
{
    url_http_t *hf;
    url_t *us;
    CURL *cu;
    int st;

    if(strcmp(mode, "r"))
	return NULL;

    /* Check cache */
    if(tchash_find(file_cache, url, -1, &hf) == 0){
	hf->cache = fopen(hf->private->cachename, "r");
	goto alloc_out;
    }

    hf = calloc(1, sizeof(url_http_t));
    hf->private = calloc(1, sizeof(http_private_t));

    if(!(cu = curl_easy_init()))
	return NULL;

    curl_easy_setopt(cu, CURLOPT_URL, url);
    curl_easy_setopt(cu, CURLOPT_NOSIGNAL, 0);
    curl_easy_setopt(cu, CURLOPT_FOLLOWLOCATION, 1);
    curl_easy_setopt(cu, CURLOPT_WRITEFUNCTION, http_recv);
    curl_easy_setopt(cu, CURLOPT_WRITEDATA, hf);
    curl_easy_setopt(cu, CURLOPT_HEADERFUNCTION, http_hdr);
    curl_easy_setopt(cu, CURLOPT_WRITEHEADER, hf);

    hf->private->curl = cu;

    /* Create and open cache file */
    hf->private->position = 0;
    hf->private->cachename = tempnam(url_http_conf_cache_dir, "http");
    hf->private->cachefd = open(hf->private->cachename,
				O_WRONLY | O_CREAT | O_TRUNC, 0600);
    hf->cache = fopen(hf->private->cachename, "r");
    pthread_cond_init(&hf->private->rc, NULL);
    pthread_mutex_init(&hf->private->mtx, NULL);

    /* Start reading */
    pthread_create(&hf->private->read_th, NULL, http_fetch, hf);

    pthread_mutex_lock(&hf->private->mtx);
    while(!(st = http_status(hf)))
	pthread_cond_wait(&hf->private->rc, &hf->private->mtx);
    pthread_mutex_unlock(&hf->private->mtx);

    if(st < 0){
	http_free(hf);
	return NULL;
    }

    /* Add to cache hash */
    tchash_replace(file_cache, url, -1, hf, NULL);

alloc_out:
    us = tcallocz(sizeof(*us));
    us->read = http_read;
    us->write = NULL;
    us->seek = http_seek;
    us->tell = http_tell;
    us->close = http_close;
    us->private = calloc(1, sizeof(url_http_t));
    *(url_http_t *)us->private = *hf;
    us->flags = URL_FLAG_STREAMED;
    us->size = hf->size;

    return us;
}
