/**
    Copyright (C) 2001-2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <jpeglib.h>
#include <jerror.h>
#include <jpeg_tc2.h>
#include <jpegmod.h>

typedef struct {
    struct jpeg_source_mgr pub;	/* public fields */
    url_t *u;
    JOCTET * buffer;		/* start of buffer */
    boolean start_of_file;	/* have we gotten any data yet? */
} io_src_mgr;

#define INPUT_BUF_SIZE  4096

/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

static void
init_source (j_decompress_ptr cinfo)
{
    io_src_mgr *src = (io_src_mgr *) cinfo->src;

    /* We reset the empty-input-file flag for each image,
     * but we don't clear the input buffer.
     * This is correct behavior for reading a series of images from one source.
     */

    src->start_of_file = TRUE;
}

static boolean
fill_input_buffer (j_decompress_ptr cinfo)
{
    io_src_mgr *src = (io_src_mgr *) cinfo->src;
    size_t nbytes;

    nbytes = src->u->read(src->buffer, 1, INPUT_BUF_SIZE, src->u);

    if (nbytes <= 0) {
	if (src->start_of_file)	/* Treat empty input file as fatal error */
	    ERREXIT(cinfo, JERR_INPUT_EMPTY);
	WARNMS(cinfo, JWRN_JPEG_EOF);
	/* Insert a fake EOI marker */
	src->buffer[0] = (JOCTET) 0xFF;
	src->buffer[1] = (JOCTET) JPEG_EOI;
	nbytes = 2;
    }

    src->pub.next_input_byte = src->buffer;
    src->pub.bytes_in_buffer = nbytes;
    src->start_of_file = FALSE;

    return TRUE;
}

static void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
    io_src_mgr *src = (io_src_mgr *) cinfo->src;

    if (num_bytes > 0) {
	while (num_bytes > (long) src->pub.bytes_in_buffer) {
	    num_bytes -= (long) src->pub.bytes_in_buffer;
	    (void) fill_input_buffer(cinfo);
	    /* note we assume that fill_input_buffer will never return FALSE,
	     * so suspension need not be handled.
	     */
	}
	src->pub.next_input_byte += (size_t) num_bytes;
	src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
}

static void
term_source (j_decompress_ptr cinfo)
{
    /* no work necessary here */
}

extern void
io_src (j_decompress_ptr cinfo, url_t *u)
{
    io_src_mgr *src;

    if (cinfo->src == NULL) {
	cinfo->src = cinfo->mem->alloc_small((j_common_ptr) cinfo,
					     JPOOL_PERMANENT,
					     sizeof(io_src_mgr));
	src = (io_src_mgr *) cinfo->src;
	src->buffer = cinfo->mem->alloc_small((j_common_ptr) cinfo,
					      JPOOL_PERMANENT,
					      INPUT_BUF_SIZE * sizeof(JOCTET));
    }

    src = (io_src_mgr *) cinfo->src;
    src->pub.init_source = init_source;
    src->pub.fill_input_buffer = fill_input_buffer;
    src->pub.skip_input_data = skip_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart;
    src->pub.term_source = term_source;
    src->u = u;
    src->pub.bytes_in_buffer = 0;
    src->pub.next_input_byte = NULL; /* until buffer loaded */
}
