/**
    Copyright (C) 2001-2004  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include <stdio.h>
#include <stdlib.h>
#include <tcalloc.h>
#include <tcendian.h>
#include <png.h>
#include <png_tc2.h>

static void png_read_wrapper(png_structp png_ptr, png_bytep data,
			     png_size_t length);
static void process_info(png_structp, png_infop, image_t *);

static void
free_png(void *p)
{
    image_t *img = p;
    free(img->data[0]);
    free(img->private);
}

extern image_t *
read_png(url_t *u, image_params_t *ip)
{
    image_t *img;
    png_structp png_ptr;
    png_infop info_ptr;
    char sig[8];

    /* Check if PNG file */
    u->read(sig, 1, 8, u);
    if(png_sig_cmp(sig, 0, 8))
	return NULL;

    /* Initialize */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    png_set_sig_bytes(png_ptr, 8);
    info_ptr = png_create_info_struct(png_ptr);
    png_set_read_fn(png_ptr, u, png_read_wrapper);

    png_read_info(png_ptr, info_ptr);

    img = tcallocdz(sizeof(*img), NULL, free_png);
    if(ip)
	img->params = *ip;
    process_info(png_ptr, info_ptr, img);

    /* Read the image */
    png_read_image(png_ptr, img->private);

    /* Clean up */
    png_read_end(png_ptr, NULL);
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

    return img;
}


/* Set transformations */
static void
process_info(png_structp png_ptr, png_infop info_ptr, image_t *img)
{
    png_bytep *row_pointers, data;
    int i, bp, width, height;
    png_byte color_type;

    /* Get width and height */
    width = png_get_image_width(png_ptr, info_ptr);
    height = png_get_image_height(png_ptr, info_ptr);

    color_type = png_get_color_type(png_ptr, info_ptr);

    /* Specific color type requested */
    if(img->params.pixel_type){
	/* Convert gray to RGB */
	if((img->params.pixel_type & IMAGE_COLOR_TYPEMASK) == IMAGE_COLOR_RGB
	   && !(color_type & PNG_COLOR_MASK_COLOR))
	    png_set_gray_to_rgb(png_ptr);
	/* Convert RGB to gray */
	else if(!(img->params.pixel_type & IMAGE_COLOR_TYPEMASK) ==
		IMAGE_COLOR_GRAY &&
		(color_type & PNG_COLOR_MASK_COLOR))
	    png_set_rgb_to_gray(png_ptr, 1, 0, 0);
    } else {
	if(color_type & PNG_COLOR_MASK_COLOR)
	    img->params.pixel_type = IMAGE_COLOR_RGB;
	else
	    img->params.pixel_type = IMAGE_COLOR_GRAY;
	if(color_type & PNG_COLOR_MASK_ALPHA)
	    img->params.pixel_type |= IMAGE_COLOR_ALPHA;
    }

    /* Add alpha to image without it */
    if((img->params.pixel_type & IMAGE_COLOR_ALPHA) &&
       !(color_type & PNG_COLOR_MASK_ALPHA)){
	png_uint_32 filler = 255;
	int flag = PNG_FILLER_AFTER;
	png_set_filler(png_ptr, filler, flag);
    }

#if TCENDIAN == TCENDIAN_LITTLE
    /* Store pixels as BGR */
    png_set_bgr(png_ptr);
#endif

    png_set_expand(png_ptr);
    png_set_interlace_handling(png_ptr);

    png_read_update_info(png_ptr, info_ptr);

    /* Compute bytes per pixel */
    bp = 1;
    if((img->params.pixel_type & IMAGE_COLOR_TYPEMASK) == IMAGE_COLOR_RGB)
	bp += 2;
    if(img->params.pixel_type & IMAGE_COLOR_ALPHA)
	bp++;
    img->params.pixelsize = bp;

    img->params.mask[0] = 0xff0000;
    img->params.mask[1] = 0xff00;
    img->params.mask[2] = 0xff;
    img->params.mask[3] = 0xff000000;

    /* Allocate memory */
    data = (png_bytep) malloc(width * height * bp);
    row_pointers = (png_bytep *) calloc(height, sizeof(png_bytep));
    for(i = 0; i < height; i++)
	row_pointers[i] = data + i * width * bp;

    img->params.width[0] = width;
    img->params.height[0] = height;
    img->data[0] = data;
    img->linesize[0] = width * bp;
    img->private = row_pointers;
}

static void
png_read_wrapper(png_structp png_ptr, png_bytep data, png_size_t length){
    url_t *u = png_get_io_ptr(png_ptr);
    u->read(data, 1, length, u);
}
