/**
    Copyright (C) 2001, 2002  Michael Ahlberg, Måns Rullgård

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
**/

#include "commands.h"
#include "tcshell.h"
#include "tclist.h"
#include <tcstring.h>
#include <stdlib.h>
#include <stdio.h>
#include <tctree.h>
#include <unistd.h>
#include <tcdirent.h>
#include <alloca.h>
#include <tctypes.h>

char delimiters[]=" ";

int cmd_load(char*);
int cmd_unload(char*);
int cmd_exit(char*);
int cmd_alias(char*);
int cmd_unalias(char*);
int cmd_lsmod(char*);
int cmd_lsdep(char*);
int cmd_ls(char*);
int cmd_cd(char*);
int cmd_system(char*);

tclist_t *cmd_list, *o_cmd_list;
tctree_t *alias_tree;

typedef struct {
    char *name;
    char *alias;
} tcsh_alias;

command builtin_cmd[13]={
    {"load", cmd_load},
    {"unload", cmd_unload},
    {"quit", cmd_exit},
    {"exit", cmd_exit},
    {"alias", cmd_alias},
    {"unalias", cmd_unalias},
    {"lsmod", cmd_lsmod},
    {"lsdep", cmd_lsdep},
    {"ls", cmd_ls},
    {"cd", cmd_cd},
    {"system", cmd_system},
    {NULL, NULL}
};

/* Compare fn. for tsearch */
int at_cmp(const void *a1, const void *a2){
    return strcmp(((tcsh_alias*)a1)->name, ((tcsh_alias*)a2)->name);
}

/* Initializes list of commands */
extern int init_cmd_list(void)
{
    int i;
    alias_tree = tctree_new(TC_LOCK_SLOPPY, at_cmp, 0);
    cmd_list=tclist_new(TC_LOCK_SLOPPY);
    o_cmd_list=tclist_new(TC_LOCK_SLOPPY);
    for(i=0;builtin_cmd[i].name!=NULL;i++){
	tclist_push(cmd_list, &builtin_cmd[i]);
    }
    
    return 0;
}

extern int uninit_cmd_list(void)
{
    return 0;
}

static int cmd_cmp_fn(const void *a, const void *b)
{
    command *a1 = (command*)a;
    command *b1 = (command*)b;
    return(strcmp(a1->name, b1->name));
}

/* Register command */
extern int shell_reg_cmd(command *cmd)
{
    command *o_cmd=NULL;

    tclist_find(cmd_list, cmd, o_cmd, cmd_cmp_fn);

    if(o_cmd == NULL) {
	tclist_push(cmd_list, cmd);
    } else {
	tclist_delete(cmd_list, o_cmd, cmd_cmp_fn, NULL);
	tclist_push(cmd_list, cmd);
	tclist_push(o_cmd_list, o_cmd);       
    }

    return 0;
}

static void
free_cmd(void *p)
{
    command *cm = p;
    free(cm->name);
    free(cm);
}

/* Unregister command */
extern int shell_unreg_cmd(command *cmd)
{
    command *o_cmd=NULL;
    
    tclist_delete(cmd_list, cmd, cmd_cmp_fn, free_cmd);

    tclist_find(o_cmd_list, cmd, &o_cmd, cmd_cmp_fn);
    if(o_cmd != NULL) {
	tclist_delete(o_cmd_list, o_cmd, cmd_cmp_fn, NULL);
	tclist_push(cmd_list, o_cmd);
    }

    return 0;
}

/* Removes whitespaces from beginning and end of string */
char *strip_white(char *in){
    char *out=in;
    int i;

    for(;;){
	if(out[0]==0)
	    break;
	if(out[0]==' ')
	    out++;
	else
	    break;
    }

    i = strlen(out);
    i--;
    while(out[i]==' '){
	out[i]=0;
	i--;
    }

    return out;
}

/* Expand alias */
char *alias_expand(char *in)
{
    char *out, *tmp, *t;
    tcsh_alias b,*c;

/*     printf("in = %s\n", in); */
    tmp = strdupa(in);
    out="";

    if((t=strsep(&tmp," "))!=NULL){
	if(strcmp(t, "unalias")==0){
	    return(strip_white(strdup(in)));
	}
	b.name=strdupa(t);
	if(tctree_find(alias_tree, &b, &c) == 0){
	    t = c->alias;
	}
/* 	printf("t = %s\n",t); */
	if(tmp==NULL){
	    tmp="";
	}
	out = malloc(strlen(t) + strlen(tmp) + 2);
	sprintf(out, "%s %s", t, tmp);
    }

/*     printf("out = %s\n", out); */
    return(strip_white(strdup(out)));
}

int strcmd_cmp(const void *p1, const void *p2)
{
    char *c = (char *) p1;
    command *d = (command *) p2;
    return strcmp(c, d->name);
}

/* Runs a command if possible */
extern int run_cmd(char *in)
{
    char *tmp, *cmd, *opt;
    command *cmd_s=NULL;

    tmp=strdupa(in);

    tmp = strip_white(tmp);

    opt = cmd = alias_expand(tmp);

    for(;opt[0]!=0;){
	if(opt[0]==' ') {
	    opt[0]=0;
	    opt++;
	    break;
	}
	opt++;
    }

    cmd = strip_white(cmd);
    opt = strip_white(opt);

    tclist_find(cmd_list, cmd, &cmd_s, strcmd_cmp);
    if(cmd_s!=NULL){
	cmd_s->cmd_fn(opt);
    }
    return 0;
}

/* Load a module (with options) */
int cmd_load(char *opt)
{
    char *mod = strdup(opt);
    char *p = strchr(mod, ':');
    char *options;

    if(p != NULL){
        *p++ = 0;
        options = strip_white(p);
    } else {
        options = NULL;
    }

    tc2_request(TC2_LOAD_MODULE, 1, mod, options);

    free(mod);

    return 0;
}

/* Unload a module */
int cmd_unload(char *opt)
{
    char *c = strdup(opt);
    tc2_request(TC2_UNLOAD_MODULE, 0, c);

    return 0;
}

/* Exit shell */
int cmd_exit(char *opt)
{
    tc2_interface **mods;
    int i, mcount;

    tc2_request(TC2_GET_MODULES, 1, &mods);
    mcount=0;

    for(i=0; mods[i]!=NULL; i++){
	if(mods[i]->loaded!=NULL){
	    mcount++;
	}
    }

    if(mcount>1) {
	char ret[10];
	printf("%d modules still loaded. Do you want to unload them? (y/n/c) ",
	       mcount-1);
	fflush(stdout);
	fgets(ret, 10, stdin);
	if((ret[0] & ~0x20) =='Y'){
	    	tc2_request(TC2_UNLOAD_ALL, 1);
		pthread_exit(NULL);
	} else if((ret[0] & ~0x20) =='N'){
	    tc2_request(TC2_UNLOAD_MODULE, 0, strdup("Shell"));
	    pthread_exit(NULL);
	}
    } else {
	tc2_request(TC2_UNLOAD_MODULE, 0, strdup("Shell"));
	pthread_exit(NULL);
    }
    return 0;
}

/* Create alias */
int cmd_alias(char *opt)
{
    char *name, *alias;
    tcsh_alias *a;
    command *cmd;
	
    a = malloc(sizeof(tcsh_alias));
    cmd = malloc(sizeof(command));
    
    name = alias = opt;
    for(;alias[0]!=0;){
	if(alias[0]==' '){
	    alias[0]=0;
	    alias++;
	    break;
	}
	alias++;
    }

    a->name=strdup(name);
    a->alias=strdup(alias);
    cmd->name=a->name;
    cmd->cmd_fn=NULL;

    tctree_replace(alias_tree, a, NULL);
    tclist_push(cmd_list, cmd);  
    return 0;
}

/* Remove an alias */
int cmd_unalias(char *opt)
{
    tcsh_alias a;
    command cmd;

    a.name=strdup(opt);
    a.alias=NULL;
    cmd.name=a.name;
    cmd.cmd_fn=NULL;

    tctree_delete(alias_tree, &a, NULL);
    tclist_delete(cmd_list, &cmd, cmd_cmp_fn, NULL);
    
    return 0;
}

/* List loaded modules */
int cmd_lsmod(char *opt)
{
    tc2_module **mods = NULL;
    int i, name_len = 7;
    char *fmt;

    tc2_request(TC2_GET_MODULES, 1, &mods);

    for(i=0; mods[i]!=NULL; i++){
	if(mods[i]->loaded!=0){
	    if(name_len < strlen(mods[i]->module->name)){
		name_len = strlen(mods[i]->module->name);
	    }
	}
    }

    /* FIXME: Find right size to allocate. */
    fmt = alloca(80);
    snprintf(fmt, 80, "%%-%ds Version Used by\n", name_len);
    printf(fmt, "Module");

    snprintf(fmt, 80, "%%-%ds (%%d.%%d.%%d) %%7d %%s\n", 
	     name_len);

    for(i=0; mods[i]!=NULL; i++){
	if(mods[i]->loaded!=0){
	    printf(fmt,
		   mods[i]->module->name,
		   mods[i]->module->version >> 16,
		   (mods[i]->module->version >> 8) & 0xFF,
		   mods[i]->module->version & 0xFF,
		   tclist_items(mods[i]->dep_list),
		   (mods[i]->autoclean? "[autoclean]":""));
	}
    }
    return 0;
}

/* List module dependencies */
int cmd_lsdep(char *opt)
{
    tc2_module *mod = NULL;
    tclist_item_t *li = NULL;
    char *dep;

    tc2_request(TC2_GET_MODULE, 1, opt, &mod);

    if(mod == NULL) {
	return 0;
    }

    printf("Dependecylist for: %s\n", mod->module->name);
 
    while((dep = tclist_next(mod->dep_list, &li)) != NULL) {    
	char *c = strdupa(dep);
	char *d = strchr(c, ':');
	tc2_module *dep_m;

	if(d != NULL){
	    *d = 0;
	}
	
	tc2_request(TC2_GET_MODULE, 1, c, &dep_m);

	if(dep_m == NULL) {
	    continue;
	}

	printf("   %s\n", dep_m->module->name);
    }

    return 0;
}

char *getdir(char *dir)
{
    char *buf, *ret;

    if(dir[0]=='/'){
	return(dir);
    } else {
	size_t s = 32;
	buf = malloc(s);
	while(getcwd(buf, s) == NULL)
	    buf = realloc(buf, s += 32);

	ret = malloc(strlen(buf) + strlen(dir) + 2);
	sprintf(ret,"%s/%s",buf,dir);
	free(buf);
	return(ret);
    }
}

static int sel_d_ent (const struct dirent *unused)
{
    return 1;
}
     
/* List files in directory */
int cmd_ls(char *opt)
{
    char *buf;
    struct dirent **eps;
    int n;

    buf = getdir(opt);

    n = scandir (buf, &eps, sel_d_ent, alphasort);
    if (n >= 0)
    {
	int cnt;
	for (cnt = 0; cnt < n; ++cnt)
	    printf("%s\n", eps[cnt]->d_name);
    }

    return 0;
}

/* Change directory */
int cmd_cd(char *opt)
{
    char *buf;

    buf = getdir(opt);
    chdir(buf);

    return 0;
}

/* System command */
int cmd_system(char *opt)
{
    system(opt);
    return 0;
}
